library nui_log;

import 'dart:convert';
import 'package:flutter_appcenter_bundle/flutter_appcenter_bundle.dart';
import 'package:nui_core/code/nui_device_util.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_database/nui_database.dart';
import 'package:nui_log/nui_log_database.dart';
import 'package:nui_log/nui_log_json.dart';
import 'package:nui_log/nui_log_models.dart';
import 'package:nui_web/nui_web.dart';
import 'nui_log_database.dart';
export 'package:nui_log/nui_log_database.dart';
export 'package:nui_log/nui_log_json.dart';
export 'package:nui_log/nui_log_models.dart';

typedef NUILogValueGetter = Future<dynamic> Function();

abstract class NUILogInterface{
  Future<NUILogCheckSettingResponse> checkLogSetting();
  Future<bool> saveLogs(List<NUILogContent> logs);
  Future<bool> saveLog(NUILogContent log);
  Future<NUILogContent> saveLogFromStackTrace({StackTrace stackTrace, dynamic error});
}

class NUILog implements NUILogInterface{
  static const _WEB_MODULE = "NUILogWebModule";
  static const _MODULE = "NUILog";
  static const _PREF_SETTING_KEY = "NUILogSettingKey";

  static NUILog _instance;

  String _baseUrl;
  String _appVersion;
  String _appId;
  String _appCenteriOSKey;
  String _appCenterAndroidKey;
  bool _appCenterInitialized = false;
  NUILogValueGetter _authorization;
  NUILogValueGetter _userIdentifier;

  NUILog._builder(NUILogBuilder builder){
    _baseUrl = builder._baseUrl;
    _appVersion = builder._appVersion;
    _appId = builder._appId;
    _appCenterAndroidKey = builder._appCenterAndroidKey;
    _appCenteriOSKey = builder._appCenterIOSKey;
    _authorization = builder._authorization;
    _userIdentifier = builder._userIdentifier;

    _initializeAppCenter(builder);
    _instance = this;
  }

  Future<dynamic> get authorization async{
    if(_authorization != null) return await _authorization();
    return null;
  }

  Future<dynamic> get userIdentifier async{
    if(_userIdentifier != null) return await _userIdentifier();
    return null;
  }

  Future<NUIWeb> _initializeWeb() async{
    if(isNullOrEmpty(_baseUrl)) return null;

    final web = await NUIWebBuilder()
        .module(NUILog._WEB_MODULE)
        .baseURL(_baseUrl)
        .build();
    return web;
  }

  Future<NUIDatabase> _initializeDatabase() async{
    final db = await NUIDatabaseBuilder().model(NUILogDatabase()).build();
    return db;
  }

  Future<bool> _initializeAppCenter(NUILogBuilder builder) async{
    await AppCenter.startAsync(
      appSecretAndroid: _appCenterAndroidKey,
      appSecretIOS: _appCenteriOSKey,
      enableAnalytics: builder._enableAnalytics ?? true, // Defaults to true
      enableCrashes: builder._enableCrashes ?? true, // Defaults to true
      enableDistribute: builder._enableDistribute ?? true, // Defaults to false
      usePrivateDistributeTrack: builder._usePrivateDistributeTrack ?? false, // Defaults to false
      disableAutomaticCheckForUpdate: builder._disableAutomaticCheckForUpdate ?? false, // Defaults to false
    );
    _appCenterInitialized = true;
    return true;
  }

  static NUILog get instance{
    if(_instance == null) {
      logNUI(_MODULE, "NUILog has not been built with builder");
      return null;
    }
    return _instance;
  }

  @override
  Future<NUILogContent> saveLogFromStackTrace({StackTrace stackTrace, dynamic error, String description}) async{
    final deviceInfo = await NUIDeviceUtil.getDeviceInfo();

    var logContent = NUILogContent(
        id: randomUUID(),
        description: description,
        action: NUILogActionType.OTHER.code(),
        error: NUILogValueType.ERROR.code(),
        bundleId: _appId,
        hardwareModel: deviceInfo.model,
        osVersion: deviceInfo.version,
        happen: DateTime.now(),
        appVersion: _appVersion);

    //Extract information from exception
    if(stackTrace != null){
      logContent.module = NUINavigator.getCurrentScreenCode();
      logContent.subModule = error?.toString();
      logContent.description = stackTrace.toString();
    }

    await _configureLogGeneralInfo(logContent, deviceInfo);

    logContent.logLevel = NUILogLevel.ERROR.value();

    saveLogs([logContent]);

    //Save exception to AppCenter
    if(stackTrace != null && _appCenterInitialized == true){
      final deviceInfo = await getDeviceInfo();
      AppCenter.trackEventAsync(
          (NUINavigator.getCurrentScreenCode().isNullOrEmpty() ? "Exception" : ("Exception at " + NUINavigator.getCurrentScreenCode())) + " v$_appVersion" + " - " + deviceInfo.model, logContent.toLogMap()
      ); // with custom properties
    }
    return logContent;
  }

  @override
  Future<NUILogCheckSettingResponse> checkLogSetting() async{
    final result = await NUIWeb.get(module: _WEB_MODULE).doPOST<NUILogCheckSettingResponse>(url: "logging/setting", entity: NUILogCheckSettingResponseEntity());
    if(result.success == true){
      logNUI(_MODULE, "Check log setting successful");
      NUISharedPref.get().save(_PREF_SETTING_KEY, result.body);
      return result.body;
    }
    else{
      logNUI(_MODULE, "Check log failed with error : ${result.message}");
      return null;
    }
  }

  Future<NUILogSetting> _getSettingByLevel(int logLevel) async{
    final setting = await NUISharedPref.get().getObject(_PREF_SETTING_KEY, NUILogCheckSettingResponseEntity());
    if(setting == null) return null;

    NUILogSetting logSetting;
    for(NUILogSetting setting in setting.data){
      if(setting.logLevel == logLevel) logSetting = setting;
      break;
    }
    return logSetting;
  }

  Future _configureLogGeneralInfo(NUILogContent log, NUIDeviceInfo deviceInfo) async{
    log.id = log.id ?? randomUUID();
    log.action = log.action ?? NUILogActionType.OTHER.code();
    log.error = log.error ?? NUILogValueType.ERROR.code();
    log.bundleId = _appId;

    log.hardwareModel = deviceInfo.model;
    log.osVersion = deviceInfo.version;
    log.happen = log.happen ?? DateTime.now();
    log.appVersion = _appVersion;

    //Get user id and session token from NUIAuth
    log.userId = (await this.userIdentifier)?.toString();
    log.sessionId = (await this.authorization)?.toString();
  }

  @override
  Future<bool> saveLog(NUILogContent log) async{
    return await saveLogs([log]);
  }

  @override
  Future<bool> saveLogs(List<NUILogContent> logs) async{
    if(isNullOrEmpty(_baseUrl)) {
      logNUI(_MODULE, "No url specified for custom log storing");
      return false;
    }

    final filteredLogs = <NUILogContent>[];
    for(NUILogContent log in logs){
      if(log.logLevel == null){
        filteredLogs.add(log);
        continue;
      }

      final logSetting = await _getSettingByLevel(log.logLevel);
      if(logSetting == null){
        filteredLogs.add(log);
      }
      else{
        if(logSetting.enable == true){
          filteredLogs.add(log);
        }
      }
    }

    if(isNullOrEmpty(filteredLogs)) {
      logNUI(_MODULE, "No logs to be saved according to the settings");
      return false;
    }

    final deviceInfo = await getDeviceInfo();
    for(var element in logs){
      await _configureLogGeneralInfo(element, deviceInfo);
    }

    if(_appCenterInitialized == true) {
      logNUI(_MODULE, "Saving ${logs.length} logs to AppCenter");
      filteredLogs.forEach((element) async {
        element.logLevel = element.logLevel ?? NUILogLevel.DEBUG.value();
        final title = (NUINavigator.getCurrentScreenCode() ?? "Log") + " v$_appVersion" + " - " + deviceInfo.model;
        logNUI(_MODULE, "Saving log : $title, content : ${jsonEncode(element)}");
        AppCenter.trackEventAsync(title, element.toLogMap()); // with custom properties
      });
    }

    final request = NUILogSaveLogRequest(list: filteredLogs);
    final result = await NUIWeb.get(module: _WEB_MODULE).doPOST(url: "logging/save", bodyEnt: request);
    if(result.success == true){
      final innerSuccess = result.responseBody["success"];
      if(innerSuccess == true) {
        logNUI(_MODULE, "Save logs successful");
        return true;
      }
      else{
        return false;
      }
    }
    else{
      logNUI(_MODULE, "Failed to save logs with error : ${result.message}");
      return false;
    }
  }
}

class NUILogBuilder{
  String _baseUrl;
  String _appVersion;
  String _appId;
  String _appCenterIOSKey;
  String _appCenterAndroidKey;
  bool _enableAnalytics;
  bool _enableCrashes;
  bool _enableDistribute;
  bool _usePrivateDistributeTrack;
  bool _disableAutomaticCheckForUpdate;
  NUILogValueGetter _authorization;
  NUILogValueGetter _userIdentifier;

  NUILogBuilder baseUrl(String baseUrl){
    _baseUrl = baseUrl;
    return this;
  }

  NUILogBuilder authorization(NUILogValueGetter authorization){
    this._authorization = authorization;
    return this;
  }

  NUILogBuilder userIdentifier(NUILogValueGetter userId){
    this._userIdentifier = userId;
    return this;
  }

  NUILogBuilder appCenteriOSKey(String key){
    _appCenterIOSKey = key;
    return this;
  }

  NUILogBuilder appCenterAndroidKey(String key){
    _appCenterAndroidKey = key;
    return this;
  }

  NUILogBuilder appVersion(String appVersion){
    _appVersion = appVersion;
    return this;
  }

  NUILogBuilder appId(String appId){
    _appId = appId;
    return this;
  }

  NUILogBuilder enableAnalytics(bool enable){
    _enableAnalytics = enable;
    return this;
  }

  NUILogBuilder enableCrashes(bool enable){
    _enableCrashes = enable;
    return this;
  }

  NUILogBuilder enableDistribute(bool enable){
    _enableDistribute = enable;
    return this;
  }

  NUILogBuilder usePrivateDistributeTrack(bool use){
    _usePrivateDistributeTrack = use;
    return this;
  }

  NUILogBuilder disableAutomaticCheckForUpdate(bool disable){
    _disableAutomaticCheckForUpdate = disable;
    return this;
  }

  Future<NUILog> build() async{
    final log = NUILog._builder(this);
    final web = log._initializeWeb();
    final db = log._initializeDatabase();
    final pref = NUISharedPrefBuilder().build();

    await web;
    await db;
    await pref;
    return NUILog._builder(this);
  }
}
